<?php

namespace Drupal\migrate_sandbox\Form;

use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\migrate\MigrateExecutable;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate_sandbox\MigrateSandboxExecutable;
use Drupal\migrate_sandbox\MigrateSandboxMessage;
use Drupal\migrate_sandbox\SandboxMigration;

/**
 * Configure Migrate Sandbox settings for this site.
 */
class MigrateSandboxForm extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'migrate_sandbox_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'migrate_sandbox.settings',
      'migrate_sandbox.processed',
      'migrate_sandbox.latest',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $form['#attached']['library'][] = 'migrate_sandbox/migrate_sandbox';
    $form['#prefix'] = '<div id="js-migrate-sandbox-wrapper">';
    $form['#suffix'] = '</div>';
    $form['prepopulated'] = [
      '#title' => $this->t('Select Starter Config'),
      '#description' => $this->t('Almost all process plugins provided by the core Migrate module and the contributed Migrate Plus module have prepopulated examples available.'),
      '#type' => 'select',
      '#options' => $this->getSourceOptions(),
      '#prefix' => '<div id="edit-prepopulated-wrapper">',
      '#suffix' => '</div>',
      '#ajax' => [
        'callback' => '::usePrepopulatedSource',
        'wrapper' => 'js-migrate-sandbox-wrapper',
      ],
    ];
    $form['intro'] = [
      '#type' => 'details',
      '#title' => $this->t('Help and Disclaimers'),
      0 => [
        '#markup' => '<h3>What is Migrate Sandbox?</h3><p>The purpose of Migrate Sandbox is to provide developers with a place to mess around with process plugins.
                        To that end, example source data and pipelines have been provided for almost all process plugins defined
                        in the core Migrate module and the contributed Migrate Plus module. You can also
                        use the sandbox to debug your custom process plugins or other contributed plugins.</p>',
      ],
      1 => [
        '#markup' => '<p><h3>Staying in the Sandbox</h3>Every attempt has been made to make this as much like a true sandbox as possible. The sandbox migration uses
                       a custom id_map that does not save any data to the db. And instead of logging errors to the db in migrate_message tables,
                       errors will be output using on-screen messages.</p>
                       <p>Though nothing goes directly into the database, the latest configuration
                       and the corresponding results get saved as configuration to <code>migrate_sandbox.latest</code> and
                       <code>migrate_sandbox.processed</code>, respectively. There is always an option to select
                       <code>migrate_sandbox.latest</code> as your starter config.</p>',
      ],
      2 => [
        '#markup' => '<h3>Sandbox Escape Warnings</h3><p>There are a number of process plugins from Migrate and Migrate Plus
                        that may result in side-effects outside of the sandbox. If you include any of the following process plugins
                        in your sandbox pipeline, there will be a warning above the Save & Run button.</p>
                        <ul>
                          <li>callback</li>
                          <li>download</li>
                          <li>file_copy</li>
                          <li>migration_lookup</li>
                          <li>entity_generate (Migrate Plus)</li>
                          <li>file_blob (Migrate Plus)</li>
                          <li>dom_entity_generate (Migrate Plus)</li>
                          <li>service (Migrate Plus)</li>
                        </ul>
                      <p>You may still use these plugins in Migrate Sandbox, but you should be aware that there could be
                      unintended consequences.</p>',
      ],
      3 => [
        '#markup' => '<h3>What this is NOT for</h3><p/>Migrate Sandbox is not intended for production sites for any reason.
                      It is not intended to run real migration. This is not designed for debugging source plugins or destination plugins.
                      It is not intended for untrusted users.</p>',
      ],
    ];

    $form['source_wrapper'] = [
      '#type' => 'fieldset',
      '#title' => 'source',
      '#tree' => FALSE,
    ];
    if (isset($this->config('migrate_sandbox.latest')->get('source.data_rows')[0])) {
      $data_row_array = $this->config('migrate_sandbox.latest')->get('source.data_rows')[0];
      unset($data_row_array['migrate_sandbox_dummy_source_id']);
      $data_row = Yaml::encode($data_row_array);
    }
    else {
      $data_row = '';
    }
    $form['source_wrapper']['data_rows'] = [
      '#type' => 'textarea',
      '#title' => 'data_rows[0]:',
      '#description' => $this->t('You can only enter one row. It can have lots of properties. But just one row.'),
      '#attributes' => ['data-yaml-editor' => 'true'],
      '#default_value' => $data_row,
    ];
    $form['source_wrapper']['preview'] = [
      '#type' => 'details',
      '#title' => $this->t('View source row as array'),
      'button' => [
        '#type' => 'button',
        '#value' => $this->t('Update'),
        '#ajax' => [
          'callback' => '::updateSourceArrayPreview',
          'wrapper' => 'source-array-preview-wrapper',
        ],
      ],
      'preview' => [
        '#prefix' => '<div id="source-array-preview-wrapper">',
        '#suffix' => '</div>',
      ],
    ];
    $form['source_wrapper']['constants'] = [
      '#type' => 'textarea',
      '#title' => 'constants:',
      '#attributes' => ['data-yaml-editor' => 'true'],
      '#default_value' => $this->config('migrate_sandbox.latest')->get('source.constants') ? Yaml::encode($this->config('migrate_sandbox.latest')->get('source.constants')) : '',
    ];
    $form['process_wrapper'] = [
      '#type' => 'fieldset',
      '#title' => 'process',
      '#tree' => FALSE,
    ];
    $form['process_wrapper']['process'] = [
      '#type' => 'textarea',
      '#title' => 'process:',
      '#attributes' => ['data-yaml-editor' => 'true'],
      '#default_value' => $this->config('migrate_sandbox.latest')->get('process') ? Yaml::encode($this->config('migrate_sandbox.latest')->get('process')) : '',
    ];
    $form['results'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Results'),
      '#weight' => 1000,
      '#prefix' => '<div id="results-wrapper">',
      '#suffix' => '</div>',
      '#tree' => FALSE,
      'format' => [
        '#type' => 'radios',
        '#title' => $this->t('Show latest results as'),
        '#options' => [
          'yaml' => 'YAML',
          'array' => 'Array',
        ],
        '#default_value' => 'yaml',
      ],
      'yaml' => [
        '#type' => 'container',
        '#states' => [
          'visible' => [
            ':input[name="format"]' => ['value' => 'yaml'],
          ],
        ],
      ],
      'array' => [
        '#type' => 'container',
        '#states' => [
          'visible' => [
            ':input[name="format"]' => ['value' => 'array'],
          ],
        ],
      ],
    ];
    $form['actions']['submit']['#value'] = $this->t('Save & Run');
    $form['actions']['submit']['#ajax'] = [
      'callback' => '::ajaxSubmit',
      'wrapper' => 'results-wrapper',
    ];
    $form['actions']['submit']['#type'] = 'button';
    $form['warnings'] = [
      '#type' => 'html_tag',
      '#tag' => 'div',
      '#attached' => [
        'drupalSettings' => [
          'migrate_sandbox_warnings' => [
            'download',
            'file_blob',
            'file_copy',
            'migration_lookup',
            'entity_generate',
            'callback',
            'service',
          ],
        ],
      ],
      '#value' => '
         <h3 class="warning-header hidden">☞ Sandbox Escape Warning</h3>
         <div class="messages messages--warning hidden" data-plugin="download">The <code>download</code> process plugin will result in files being created outside of the sandbox.</div>
         <div class="messages messages--warning hidden" data-plugin="file_blob">The <code>file_blob</code> process plugin will result in files being created outside of the sandbox.</div>
         <div class="messages messages--warning hidden" data-plugin="file_copy">The <code>file_copy</code> process plugin will result in files being created outside of the sandbox.</div>
         <div class="messages messages--warning hidden" data-plugin="migration_lookup">The <code>migration_lookup</code> process plugin can result in Drupal entities being created and saved outside of the sandbox. Use the <code>no_stub</code> configuration to prevent this.</code></div>
         <div class="messages messages--warning hidden" data-plugin="entity_generate">The <code>entity_generate</code> process plugin will result in Drupal entities being created and saved outside of the sandbox.</div>
         <div class="messages messages--warning hidden" data-plugin="callback">The <code>callback</code> process plugin can cause side-effects outside of the sandbox if the callable is not a pure function.</div>
         <div class="messages messages--warning hidden" data-plugin="service">The <code>service</code> process plugin can cause side-effects outside of the sandbox if the method is not a pure function.</div>
         <div class="messages messages--warning hidden" data-plugin="dom_migration_lookup">The <code>dom_migration_lookup</code> process plugin can result in Drupal entities being created and saved outside of the sandbox. Use the <code>no_stub</code> configuration to prevent this.</code></div>
         ',
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->getTriggeringElement()['#ajax']['callback'] == '::updateSourceArrayPreview') {
      $yaml_fields = [
        'data_rows',
      ];
    }
    else {
      $yaml_fields = [
        'data_rows',
        'constants',
        'process',
      ];
    }
    foreach ($yaml_fields as $field) {
      try {
        Yaml::decode($form_state->getValue($field));
      }
      catch (\Exception $e) {
        $form_state->setErrorByName($field, $field . ': ' . $e->getMessage());
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function ajaxSubmit(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getErrors()) {
      $id = 'migrate_sandbox';
      $label = 'Migrate Sandbox';
      $source_id = [
        'migrate_sandbox_dummy_source_id' => '123',
      ];
      $ids = [
        'migrate_sandbox_dummy_source_id' => [
          'type' => 'integer',
        ],
      ];
      $source = [
        'plugin' => 'embedded_data',
        'data_rows' => [(Yaml::decode($form_state->getValue('data_rows')) ?? []) + $source_id],
        'constants' => Yaml::decode($form_state->getValue('constants')),
        'ids' => $ids,
      ];
      $process = Yaml::decode($form_state->getValue('process'));
      $destination = [
        'plugin' => 'config',
        'config_name' => 'migrate_sandbox.processed',
      ];
      $this->config('migrate_sandbox.latest')
        ->setData([
          'id' => $id,
          'label' => $label,
          'source' => $source,
          'process' => $process,
          'destination' => $destination,
        ])
        ->save();
      $results = $this->runMigration();
      $form['results']['value']['#value'] = $results;
      $form_state->setValue('results_array', $results);
      $form['results']['yaml'][0]['#markup'] = '<pre>' . Yaml::encode($results) . '</pre>';
      $form['results']['array'][0]['#markup'] = '<pre>' . print_r($results, TRUE) . '</pre>';
      $form_state->setRebuild();
    }
    return $form['results'];
  }

  /**
   * {@inheritdoc}
   */
  protected function runMigration() {
    $definition = $this->config('migrate_sandbox.latest')->getRawData();
    $migration = SandboxMigration::create(\Drupal::getContainer(), [], 'migrate_sandbox', $definition);
    $migration->setStatus(MigrationInterface::STATUS_IDLE);
    $executable = new MigrateExecutable($migration, new MigrateSandboxMessage());
    try {
      $this->config('migrate_sandbox.processed')->setData([])->save();
      $executable->import();
      return $this->config('migrate_sandbox.processed')->get();
    }
    catch (\Throwable $e) {
      \Drupal::messenger()->addError($e->getMessage());
      return [];
    }
  }

  /**
   * Get array of source options.
   */
  protected function getSourceOptions() {
    $sources = $this->config('migrate_sandbox.settings')->getRawData()['sources'];
    $options = ['' => 'migrate_sandbox.latest'];
    foreach ($sources as $source) {
      $options[$source['name']] = $source['name'];
    }
    return $options;
  }

  /**
   * Handles using prepopulated source data.
   */
  public function usePrepopulatedSource($form, FormStateInterface $form_state) {
    if ($source_name = $form_state->getValue('prepopulated')) {
      $sources = $this->config('migrate_sandbox.settings')->getRawData()['sources'];
      $prepopulated = [];
      foreach ($sources as $source) {
        if ($source['name'] === $form_state->getValue('prepopulated')) {
          $prepopulated = $source;
          break;
        }
      }
    }
    else {
      $prepopulated = $this->config('migrate_sandbox.latest')->getRawData();
      if (empty($prepopulated)) {
        $prepopulated = [
          'source' => [
            'data_rows' => NULL,
            'constants' => NULL,
          ],
          'process' => NULL,
        ];
      }
    }
    if (isset($prepopulated['source']['data_rows'][0])) {
      unset($prepopulated['source']['data_rows'][0]['migrate_sandbox_dummy_source_id']);
      $form['source_wrapper']['data_rows']['#value'] = Yaml::encode($prepopulated['source']['data_rows'][0]);
    }
    else {
      $form['source_wrapper']['data_rows']['#value'] = '';
    }
    $form['source_wrapper']['constants']['#value'] = isset($prepopulated['source']['constants']) ? Yaml::encode($prepopulated['source']['constants']) : '';
    $form['process_wrapper']['process']['#value'] = isset($prepopulated['process']) ? Yaml::encode($prepopulated['process']) : '';
    $form['source_wrapper']['preview']['preview']['#markup'] = NULL;
    $form['results']['yaml'][0]['#markup'] = NULL;
    $form['results']['array'][0]['#markup'] = NULL;
    $form_state->setRebuild();
    return $form;
  }

  /**
   * Updates source array preview.
   */
  public function updateSourceArrayPreview($form, FormStateInterface $form_state) {
    try {
      $data_row_array = Yaml::decode($form_state->getValue('data_rows'));
      $preview = print_r($data_row_array, TRUE);
      $form['source_wrapper']['preview']['preview']['#markup'] = '<pre>' . $preview . '</pre>';
    }
    catch (\Exception $e) {
      // Message will be shown due to validation function.
    }
    $form_state->setRebuild();
    return $form['source_wrapper']['preview']['preview'];
  }

}
