<?php

namespace Drupal\Tests\migrate_sandbox\FunctionalJavascript;

use Drupal\FunctionalJavascriptTests\WebDriverTestBase;

/**
 * Tests migrate_sandbox.
 *
 * @group migrate_sandbox
 */
class MigrateSandboxTest extends WebDriverTestBase {

  /**
   * Modules to install.
   *
   * @var array
   */
  protected static $modules = [
    'migrate',
    'migrate_sandbox',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * The schema is still a @todo.
   *
   * @var bool
   */
  protected $strictConfigSchema = FALSE;

  /**
   * Tests the migrate_sandbox form.
   */
  public function testMigrateSandbox() {
    // Test as trusted user.
    $trusted_user = $this->drupalCreateUser(['access migrate_sandbox']);
    $this->drupalLogin($trusted_user);
    $this->drupalGet('/admin/config/development/migrate-sandbox');
    $page = $this->getSession()->getPage();
    $assert_session = $this->assertSession();

    // Test that urlencode works.
    $page->find('css', '[name="prepopulated"]')->setValue('urlencode');
    $assert_session->waitForText("my_url_with_spaces: 'http://example.com/a url with spaces.html')");
    $page->find('css', '[name="constants"]')->setValue('hello: howdy');
    $assert_session->pageTextNotContains("new_url: 'http://example.com/a%20url%20with%20spaces.html'");
    $page->pressButton('Save & Run');
    $assert_session->waitForText("new_url: 'http://example.com/a%20url%20with%20spaces.html'");

    // Test that config was saved correctly.
    $latest = \Drupal::configFactory()->get('migrate_sandbox.latest');
    $this->assertSame('migrate_sandbox', $latest->get('id'));
    $this->assertSame('Migrate Sandbox', $latest->get('label'));
    $this->assertSame('embedded_data', $latest->get('source.plugin'));
    $this->assertSame('http://example.com/a url with spaces.html', $latest->get('source.data_rows')[0]['my_url_with_spaces']);
    $this->assertSame('123', $latest->get('source.data_rows')[0]['migrate_sandbox_dummy_source_id']);
    $this->assertSame('howdy', $latest->get('source.constants.hello'));
    $this->assertSame('urlencode', $latest->get('process.new_url.plugin'));
    $this->assertSame('my_url_with_spaces', $latest->get('process.new_url.source'));
    $this->assertSame('config', $latest->get('destination.plugin'));
    $this->assertSame('migrate_sandbox.processed', $latest->get('destination.config_name'));
    $processed = \Drupal::configFactory()->get('migrate_sandbox.processed');
    $this->assertSame('http://example.com/a%20url%20with%20spaces.html', $processed->get('new_url'));

    // That migrate map does not exist.
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_map_migrate_sandbox'));
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_message_migrate_sandbox'));

    // Test sandbox escape warnings.
    $callback_warning = $page->find('css', '[data-plugin="callback"]');
    $lookup_warning = $page->find('css', '[data-plugin="migration_lookup"]');
    $this->assertFalse($callback_warning->isVisible());
    $this->assertFalse($lookup_warning->isVisible());
    $page->find('css', '[name="process"]')->setValue('plugin: callback');
    $assert_session->waitForElementVisible('css', '[data-plugin="callback"]');
    $this->assertTrue($callback_warning->isVisible());
    $this->assertFalse($lookup_warning->isVisible());
    $page->find('css', '[name="process"]')->setValue('plugin: "migration_lookup"');
    $assert_session->waitForElementVisible('css', '[data-plugin="migration_lookup"]');
    $this->assertFalse($callback_warning->isVisible());
    $this->assertTrue($lookup_warning->isVisible());

    // Run log example to test for on-screen messages.
    $page->find('css', '[name="prepopulated"]')->setValue('log');
    $assert_session->waitForText("foo: Foo");
    $expected_log = "'example_1' value is 'Foo'";
    $assert_session->pageTextNotContains($expected_log);
    $page->pressButton('Save & Run');
    $assert_session->waitForElementVisible('css', 'div[aria-label="Error message"]');
    $assert_session->elementTextContains('css', 'div[aria-label="Error message"]', $expected_log);
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_map_migrate_sandbox'));
    $this->assertFalse(\Drupal::database()->schema()->tableExists('migrate_message_migrate_sandbox'));

    // Try submitting with invalid yaml. First re-load page to clear messages.
    $this->drupalGet('/admin/config/development/migrate-sandbox');
    $page->find('css', '[name="data_rows"]')->setValue(' foo:');
    $assert_session->elementNotExists('css', 'div[aria-label="Error message"]');
    $page->pressButton('Save & Run');
    $assert_session->waitForElementVisible('css', 'div[aria-label="Error message"]');

    // Test XSS.
    $page->find('css', '[name="data_rows"]')->setValue('xss: <script>alert("gotcha")</script>');
    $page->find('css', '[name="process"]')->setValue('xss: xss');
    $page->pressButton('Save & Run');
    $assert_session->waitForText("xss: 'alert(\"gotcha\")");
    $processed = \Drupal::configFactory()->get('migrate_sandbox.processed');
    $page->find('css', '[name="data_rows"]')->setValue('');
    $page->find('css', '[name="process"]')->setValue('');
    $page->find('css', '[name="prepopulated"]')->setValue('');
    $assert_session->waitForText('xss: \'<script>alert("gotcha")</script>\'');
    $page->pressButton('Save & Run');
    $assert_session->waitForText("xss: 'alert(\"gotcha\")");
  }

}
